from ..base import BaseDatasetLoader


class PydalaDatasetLoader(BaseDatasetLoader):
    def model_post_init(self, __context):
        super().model_post_init(__context)
        self.format = "parquet"

    def load(self, **kwargs):
        return self.to_pydala_dataset(**kwargs)



import base64
import datetime as dt
import inspect
import os
import posixpath
import urllib
import uuid
from pathlib import Path
from typing import Generator

import fsspec
import orjson
import pandas as pd
import pyarrow as pa
import pyarrow.dataset as pds
import pyarrow.parquet as pq
import requests
from fsspec import filesystem
from fsspec.implementations.cache_mapper import AbstractCacheMapper
from fsspec.implementations.cached import SimpleCacheFileSystem
from fsspec.implementations.dirfs import DirFileSystem
from fsspec.implementations.memory import MemoryFile
from fsspec.spec import AbstractFileSystem
from fsspec.utils import infer_storage_options
from loguru import logger

from .misc import convert_large_types_to_standard, run_parallel
from .polars import pl
from .storage_options import (
    AwsStorageOptions,
    AzureStorageOptions,
    GcsStorageOptions,
    GitHubStorageOptions,
    GitLabStorageOptions,
    get_storage_options,
)


class FileNameCacheMapper(AbstractCacheMapper):
    def __init__(self, directory):
        self.directory = directory

    def __call__(self, path: str) -> str:
        os.makedirs(
            posixpath.dirname(posixpath.join(self.directory, path)), exist_ok=True
        )
        return path


class MonitoredSimpleCacheFileSystem(SimpleCacheFileSystem):
    def __init__(self, **kwargs):
        # kwargs["cache_storage"] = posixpath.join(
        #    kwargs.get("cache_storage"), kwargs.get("fs").protocol[0]
        # )
        self._verbose = kwargs.get("verbose", False)
        super().__init__(**kwargs)
        self._mapper = FileNameCacheMapper(kwargs.get("cache_storage"))

    def _check_file(self, path):
        self._check_cache()
        cache_path = self._mapper(path)
        for storage in self.storage:
            fn = posixpath.join(storage, cache_path)
            if posixpath.exists(fn):
                return fn
            if self._verbose:
                logger.info(f"Downloading {self.protocol[0]}://{path}")

    # def glob(self, path):
    #    return [self._strip_protocol(path)]

    def size(self, path):
        cached_file = self._check_file(self._strip_protocol(path))
        if cached_file is None:
            return self.fs.size(path)
        else:
            return posixpath.getsize(cached_file)

    def sync(self, reload: bool = False):
        if reload:
            self.clear_cache()
        content = self.glob("**/*")
        [self.open(f).close() for f in content if self.isfile(f)]

    def __getattribute__(self, item):
        if item in {
            # new items
            "size",
            "glob",
            "sync",
            # previous
            "load_cache",
            "_open",
            "save_cache",
            "close_and_update",
            "__init__",
            "__getattribute__",
            "__reduce__",
            "_make_local_details",
            "open",
            "cat",
            "cat_file",
            "cat_ranges",
            "get",
            "read_block",
            "tail",
            "head",
            "info",
            "ls",
            "exists",
            "isfile",
            "isdir",
            "_check_file",
            "_check_cache",
            "_mkcache",
            "clear_cache",
            "clear_expired_cache",
            "pop_from_cache",
            "local_file",
            "_paths_from_path",
            "get_mapper",
            "open_many",
            "commit_many",
            "hash_name",
            "__hash__",
            "__eq__",
            "to_json",
            "to_dict",
            "cache_size",
            "pipe_file",
            "pipe",
            "start_transaction",
            "end_transaction",
        }:
            # all the methods defined in this class. Note `open` here, since
            # it calls `_open`, but is actually in superclass
            return lambda *args, **kw: getattr(type(self), item).__get__(self)(
                *args, **kw
            )
        if item in ["__reduce_ex__"]:
            raise AttributeError
        if item in ["transaction"]:
            # property
            return type(self).transaction.__get__(self)
        if item in ["_cache", "transaction_type"]:
            # class attributes
            return getattr(type(self), item)
        if item == "__class__":
            return type(self)
        d = object.__getattribute__(self, "__dict__")
        fs = d.get("fs", None)  # fs is not immediately defined
        if item in d:
            return d[item]
        elif fs is not None:
            if item in fs.__dict__:
                # attribute of instance
                return fs.__dict__[item]
            # attributed belonging to the target filesystem
            cls = type(fs)
            m = getattr(cls, item)
            if (inspect.isfunction(m) or inspect.isdatadescriptor(m)) and (
                not hasattr(m, "__self__") or m.__self__ is None
            ):
                # instance method
                return m.__get__(fs, cls)
            return m  # class method or attribute
        else:
            # attributes of the superclass, while target is being set up
            return super().__getattribute__(item)


class GitLabFileSystem(AbstractFileSystem):
    def __init__(
        self,
        project_name: str | None = None,
        project_id: str | None = None,
        access_token: str | None = None,
        branch: str = "main",
        base_url: str = "https://gitlab.com",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.project_name = project_name
        self.project_id = project_id
        self.access_token = access_token
        self.branch = branch
        self.base_url = base_url.rstrip("/")
        self._validate_init()
        if not self.project_id:
            self.project_id = self._get_project_id()

    def _validate_init(self):
        if not self.project_id and not self.project_name:
            raise ValueError("Either 'project_id' or 'project_name' must be provided")

    def _get_project_id(self):
        url = f"{self.base_url}/api/v4/projects"
        headers = {"PRIVATE-TOKEN": self.access_token}
        params = {"search": self.project_name}
        response = requests.get(url, headers=headers, params=params)

        if response.status_code == 200:
            projects = response.json()
            for project in projects:
                if project["name"] == self.project_name:
                    return project["id"]
            raise ValueError(f"Project '{self.project_name}' not found")
        else:
            response.raise_for_status()

    def _open(self, path, mode="rb", **kwargs):
        if mode != "rb":
            raise NotImplementedError("Only read mode is supported")

        url = (
            f"{self.base_url}/api/v4/projects/{self.project_id}/repository/files/"
            f"{urllib.parse.quote_plus(path)}?ref={self.branch}"
        )
        headers = {"PRIVATE-TOKEN": self.access_token}
        response = requests.get(url, headers=headers)

        if response.status_code == 200:
            file_content = base64.b64decode(response.json()["content"])
            return MemoryFile(None, None, file_content)
        else:
            response.raise_for_status()

    def _ls(self, path, detail=False, **kwargs):
        url = f"{self.base_url}/api/v4/projects/{self.project_id}/repository/tree?path={path}&ref={self.branch}"
        headers = {"PRIVATE-TOKEN": self.access_token}
        response = requests.get(url, headers=headers)

        if response.status_code == 200:
            files = response.json()
            if detail:
                return files
            else:
                return [file["name"] for file in files]
        else:
            response.raise_for_status()


try:
    fsspec.register_implementation("gitlab", GitLabFileSystem)
except ValueError as e:
    _ = e


# Original ls Methode speichern
dirfs_ls_o = DirFileSystem.ls
mscf_ls_o = MonitoredSimpleCacheFileSystem.ls


# Neue ls Methode definieren
def dir_ls_p(self, path, detail=False, **kwargs):
    return dirfs_ls_o(self, path, detail=detail, **kwargs)


def mscf_ls_p(self, path, detail=False, **kwargs):
    return mscf_ls_o(self, path, detail=detail, **kwargs)


def read_json_file(
    path, self, include_file_path: bool = False, jsonlines: bool = False
) -> dict | list[dict]:
    with self.open(path) as f:
        if jsonlines:
            data = [orjson.loads(line) for line in f.readlines()]
        else:
            data = orjson.loads(f.read())
    if include_file_path:
        return {path: data}
    return data


def _read_json(
    self,
    path: str | list[str],
    include_file_path: bool = False,
    use_threads: bool = True,
    jsonlines: bool = False,
    as_dataframe: bool = True,
    concat: bool = True,
    verbose: bool = False,
    **kwargs,
) -> dict | list[dict] | pl.DataFrame | list[pl.DataFrame]:
    """
    Read a JSON file or a list of JSON files.

    Args:
        path: (str | list[str]) Path to the JSON file(s).
        include_file_path: (bool, optional) If True, return a dictionary with the file path as key.
            Defaults to False.
        use_threads: (bool, optional) If True, read files in parallel. Defaults to True.
        jsonlines: (bool, optional) If True, read JSON lines. Defaults to False.
        as_dataframe: (bool, optional) If True, return a DataFrame. Defaults to True.
        concat: (bool, optional) If True, concatenate the DataFrames. Defaults to True.
        verbose: (bool, optional) If True, print verbose output. Defaults to False.
        **kwargs: Additional keyword arguments.

    Returns:
        (dict | list[dict] | pl.DataFrame | list[pl.DataFrame]):
            Dictionary, list of dictionaries, DataFrame or list of DataFrames.
    """
    if isinstance(path, str):
        if "**" in path:
            path = self.glob(path)
        else:
            if ".json" not in posixpath.basename(path):
                path = posixpath.join(path, "**/*.jsonl" if jsonlines else "**/*.json")
                path = self.glob(path)

    if isinstance(path, list):
        if use_threads:
            data = run_parallel(
                read_json_file,
                path,
                self=self,
                include_file_path=include_file_path,
                jsonlines=jsonlines,
                n_jobs=-1,
                backend="threading",
                verbose=verbose,
                **kwargs,
            )
        data = [
            read_json_file(
                path=p,
                self=self,
                include_file_path=include_file_path,
                jsonlines=jsonlines,
            )
            for p in path
        ]
    data = read_json_file(
        path=path, self=self, include_file_path=include_file_path, jsonlines=jsonlines
    )
    if as_dataframe:
        if not include_file_path:
            data = pl.DataFrame(data)
        else:
            data = pl.DataFrame(list(data.values())[0]).with_columns(
                pl.lit(list(data.keys())[0]).alias("file_path")
            )
        if concat:
            return pl.concat(data, how="diagonal_relaxed")
    return data


def _read_json_batches(
    self,
    path: str | list[str],
    num_batches: int = 1,
    include_file_path: bool = False,
    jsonlines: bool = False,
    as_dataframe: bool = True,
    concat: bool = True,
    use_threads: bool = True,
    verbose: bool = False,
    **kwargs,
) -> Generator[dict | list[dict] | pl.DataFrame | list[pl.DataFrame], None, None]:
    """
    Read JSON files in batches with optional parallel processing within batches.

    Args:
        path: (str | list[str]) Path to the JSON file(s).
        num_batches: (int) Number of files per batch. Defaults to 1.
        include_file_path: (bool) If True, return with file path as key.
        jsonlines: (bool) If True, read JSON lines. Defaults to False.
        as_dataframe: (bool) If True, return DataFrame. Defaults to True.
        concat: (bool) If True, concatenate batch DataFrames. Defaults to True.
        use_threads: (bool) If True, use parallel processing within batches.
        verbose: (bool) If True, print verbose output.
        **kwargs: Additional keyword arguments.

    Yields:
        Data from num_batches files as dict/DataFrame based on parameters.
    """
    # Handle path resolution
    if isinstance(path, str):
        if "**" in path:
            path = self.glob(path)
        else:
            if ".json" not in posixpath.basename(path):
                path = posixpath.join(path, "**/*.jsonl" if jsonlines else "**/*.json")
                path = self.glob(path)

    if isinstance(path, str):
        path = [path]

    # Process files in batches
    for i in range(0, len(path), num_batches):
        batch_paths = path[i : i + num_batches]

        # Read batch with optional parallelization
        if use_threads and len(batch_paths) > 1:
            batch_data = run_parallel(
                read_json_file,
                batch_paths,
                self=self,
                include_file_path=include_file_path,
                jsonlines=jsonlines,
                n_jobs=-1,
                backend="threading",
                verbose=verbose,
                **kwargs,
            )
        else:
            batch_data = [
                read_json_file(
                    path=p,
                    self=self,
                    include_file_path=include_file_path,
                    jsonlines=jsonlines,
                )
                for p in batch_paths
            ]

        if as_dataframe:
            if not include_file_path:
                batch_dfs = [pl.DataFrame(d) for d in batch_data]
            else:
                batch_dfs = [
                    pl.DataFrame(list(d.values())[0]).with_columns(
                        pl.lit(list(d.keys())[0]).alias("file_path")
                    )
                    for d in batch_data
                ]

            if concat and len(batch_dfs) > 1:
                yield pl.concat(batch_dfs, how="diagonal_relaxed")
            else:
                yield batch_dfs
        else:
            yield batch_data


def read_json(
    self,
    path: str | list[str],
    num_batches: int | None = None,
    include_file_path: bool = False,
    jsonlines: bool = False,
    as_dataframe: bool = True,
    concat: bool = True,
    use_threads: bool = True,
    verbose: bool = False,
    **kwargs,
) -> (
    dict
    | list[dict]
    | pl.DataFrame
    | list[pl.DataFrame]
    | Generator[dict | list[dict] | pl.DataFrame | list[pl.DataFrame], None, None]
):
    """
    Read a JSON file or a list of JSON files. Optionally read in batches,
    returning a generator that sequentially yields data for specified number of files.

    Args:
        path: (str | list[str]) Path to the JSON file(s).
        num_batches: (int, optional) Number of files to process in each batch. Defaults to 1.
        include_file_path: (bool) If True, return with file path as key.
        jsonlines: (bool) If True, read JSON lines. Defaults to False.
        as_dataframe: (bool) If True, return DataFrame. Defaults to True.
        concat: (bool) If True, concatenate the DataFrames. Defaults to True.
        use_threads: (bool) If True, use parallel processing within batches. Defaults to True.
        verbose: (bool) If True, print verbose output. Defaults to False.
        **kwargs: Additional keyword arguments.

    Returns:
        (dict | list[dict] | pl.DataFrame | list[pl.DataFrame]:
            Dictionary, list of dictionaries, DataFrame, list of DataFrames
    Yields:
        (dict | list[dict] | pl.DataFrame | list[pl.DataFrame]:
            Dictionary, list of dictionaries, DataFrame, list of DataFrames
    """
    if num_batches is not None:
        yield from _read_json_batches(
            self=self,
            path=path,
            num_batches=num_batches,
            include_file_path=include_file_path,
            jsonlines=jsonlines,
            as_dataframe=as_dataframe,
            concat=concat,
            use_threads=use_threads,
            verbose=verbose,
            **kwargs,
        )
    return _read_json(
        self=self,
        path=path,
        include_file_path=include_file_path,
        jsonlines=jsonlines,
        as_dataframe=as_dataframe,
        concat=concat,
        use_threads=use_threads,
        verbose=verbose,
        **kwargs,
    )


def read_csv_file(
    path, self, include_file_path: bool = False, **kwargs
) -> pl.DataFrame:
    with self.open(path) as f:
        df = pl.read_csv(f, **kwargs)
    if include_file_path:
        return df.with_columns(pl.lit(path).alias("file_path"))
    return df


def _read_csv(
    self,
    path: str | list[str],
    include_file_path: bool = False,
    use_threads: bool = True,
    concat: bool = True,
    verbose: bool = False,
    **kwargs,
) -> pl.DataFrame | list[pl.DataFrame]:
    """
    Read a CSV file or a list of CSV files into a polars DataFrame.

    Args:
        path: (str | list[str]) Path to the CSV file(s).
        include_file_path: (bool, optional) If True, return a DataFrame with a 'file_path' column.
            Defaults to False.
        use_threads: (bool, optional) If True, read files in parallel. Defaults to True.
        concat: (bool, optional) If True, concatenate the DataFrames. Defaults to True.
        verbose: (bool, optional) If True, print verbose output. Defaults to False.
        **kwargs: Additional keyword arguments.

    Returns:
        (pl.DataFrame | list[pl.DataFrame]): Polars DataFrame or list of DataFrames.
    """
    if isinstance(path, str):
        if "**" in path:
            path = self.glob(path)
        else:
            if ".csv" not in posixpath.basename(path):
                path = posixpath.join(path, "**/*.csv")
                path = self.glob(path)

    if isinstance(path, list):
        if use_threads:
            dfs = run_parallel(
                read_csv_file,
                path,
                self=self,
                include_file_path=include_file_path,
                n_jobs=-1,
                backend="threading",
                verbose=verbose,
                **kwargs,
            )
        dfs = [
            read_csv_file(p, self=self, include_file_path=include_file_path, **kwargs)
            for p in path
        ]
    dfs = read_csv_file(
        path=path, self=self, include_file_path=include_file_path, **kwargs
    )
    if concat:
        return pl.concat(dfs, how="diagonal_relaxed")
    return dfs


def _read_csv_batches(
    self,
    path: str | list[str],
    num_batches: int = 1,
    include_file_path: bool = False,
    concat: bool = True,
    use_threads: bool = True,
    verbose: bool = False,
    **kwargs,
) -> Generator[pl.DataFrame, None, None]:
    """
    Read CSV files in batches with optional parallel processing within batches.

    Args:
        path: (str | list[str]) Path to the CSV file(s).
        num_batches: (int) Number of files per batch. Defaults to 1.
        include_file_path: (bool) If True, include file_path column.
        concat: (bool) If True, concatenate batch DataFrames.
        use_threads: (bool) If True, use parallel processing within batches.
        verbose: (bool) If True, print verbose output.
        **kwargs: Additional keyword arguments.

    Yields:
        pl.DataFrame: DataFrame containing data from num_batches files.
    """
    # Handle path resolution
    if isinstance(path, str):
        if "**" in path:
            path = self.glob(path)
        else:
            if ".csv" not in posixpath.basename(path):
                path = posixpath.join(path, "**/*.csv")
                path = self.glob(path)

    # Ensure path is a list
    if isinstance(path, str):
        path = [path]

    # Process files in batches
    for i in range(0, len(path), num_batches):
        batch_paths = path[i : i + num_batches]

        # Read batch with optional parallelization
        if use_threads and len(batch_paths) > 1:
            batch_dfs = run_parallel(
                read_csv_file,
                batch_paths,
                self=self,
                include_file_path=include_file_path,
                n_jobs=-1,
                backend="threading",
                verbose=verbose,
                **kwargs,
            )
        else:
            batch_dfs = [
                read_csv_file(
                    p, self=self, include_file_path=include_file_path, **kwargs
                )
                for p in batch_paths
            ]

        if concat and len(batch_dfs) > 1:
            yield pl.concat(batch_dfs, how="diagonal_relaxed")
        else:
            yield batch_dfs


def read_csv(
    self,
    path: str | list[str],
    num_batches: int | None = None,
    include_file_path: bool = False,
    concat: bool = True,
    use_threads: bool = True,
    verbose: bool = False,
    **kwargs,
) -> (
    pl.DataFrame
    | list[pl.DataFrame]
    | Generator[pl.DataFrame | list[pl.DataFrame], None, None]
):
    """
    Read a CSV file or a list of CSV files. Optionally read in batches,
    returning a generator that sequentially yields data for specified number of files.

    Args:
        path: (str | list[str]) Path to the CSV file(s).
        num_batches: (int, optional) Number of files to process in each batch. Defaults to 1.
        include_file_path: (bool, optional) If True, include 'file_path' column.
        concat: (bool, optional) If True, concatenate the batch DataFrames. Defaults to True.
        use_threads: (bool, optional) If True, use parallel processing within batches. Defaults to True.
        verbose: (bool, optional) If True, print verbose output. Defaults to False.
        **kwargs: Additional keyword arguments.

    Returns:
        pl.DataFrame | list[pl.DataFrame]:
            DataFrame or list or DataFrames containing data from num_batches files.

    Yields:
        pl.DataFrame | list[pl.DataFrame]:
            DataFrame or list of DataFrames containing data from num_batches files.
    """
    if num_batches is not None:
        yield from _read_csv_batches(
            self=self,
            path=path,
            num_batches=num_batches,
            include_file_path=include_file_path,
            concat=concat,
            use_threads=use_threads,
            verbose=verbose,
            **kwargs,
        )
    return _read_csv(
        self=self,
        path=path,
        include_file_path=include_file_path,
        concat=concat,
        use_threads=use_threads,
        verbose=verbose,
        **kwargs,
    )


def read_parquet_file(
    path, self, include_file_path: bool = False, **kwargs
) -> pa.Table:
    table = pq.read_table(path, filesystem=self, **kwargs)
    if include_file_path:
        return table.add_column(0, "file_path", pl.Series([path] * table.num_rows))
    return table


def _read_parquet(
    self,
    path,
    include_file_path: bool = False,
    use_threads: bool = True,
    concat: bool = True,
    verbose: bool = False,
    **kwargs,
) -> pa.Table | list[pa.Table]:
    """
    Read a Parquet file or a list of Parquet files into a pyarrow Table.

    Args:
        path: (str | list[str]) Path to the Parquet file(s).
        include_file_path: (bool, optional) If True, return a Table with a 'file_path' column.
            Defaults to False.
        use_threads: (bool, optional) If True, read files in parallel. Defaults to True.
        concat: (bool, optional) If True, concatenate the Tables. Defaults to True.
        **kwargs: Additional keyword arguments.

    Returns:
        (pa.Table | list[pa.Table]): Pyarrow Table or list of Pyarrow Tables.
    """
    if not include_file_path and concat:
        return pq.read_table(path, filesystem=self, **kwargs)
    else:
        if isinstance(path, str):
            if "**" in path:
                path = self.glob(path)
            else:
                if ".parquet" not in posixpath.basename(path):
                    path = posixpath.join(path, "**/*.parquet")
                    path = self.glob(path)

        if isinstance(path, list):
            if use_threads:
                table = run_parallel(
                    read_parquet_file,
                    path,
                    self=self,
                    include_file_path=include_file_path,
                    n_jobs=-1,
                    backend="threading",
                    verbose=verbose,
                    **kwargs,
                )
            table = [
                read_parquet_file(
                    p, self=self, include_file_path=include_file_path, **kwargs
                )
                for p in path
            ]

    table = read_parquet_file(
        path=path, self=self, include_file_path=include_file_path, **kwargs
    )
    if concat:
        return pa.concat_tables(table, promote_options="permissive")
    return table


def _read_parquet_batches(
    self,
    path: str | list[str],
    num_batches: int = 1,
    include_file_path: bool = False,
    concat: bool = True,
    use_threads: bool = True,
    verbose: bool = False,
    **kwargs,
) -> Generator[pa.Table, None, None]:
    """
    Read Parquet files in batches with optional parallel processing within batches.

    Args:
        path: (str | list[str]) Path to the Parquet file(s).
        num_batches: (int) Number of files per batch. Defaults to 1.
        include_file_path: (bool) If True, include file_path column.
        concat: (bool) If True, concatenate batch tables.
        use_threads: (bool) If True, use parallel processing within batches.
        verbose: (bool) If True, print verbose output.
        **kwargs: Additional keyword arguments.

    Yields:
        pa.Table: PyArrow Table containing data from num_batches files.
    """

    # Handle path resolution
    if isinstance(path, str):
        if "**" in path:
            path = self.glob(path)
        else:
            if ".parquet" not in posixpath.basename(path):
                path = posixpath.join(path, "**/*.parquet")
                path = self.glob(path)

    if isinstance(path, str):
        path = [path]

    # Process files in batches
    for i in range(0, len(path), num_batches):
        batch_paths = path[i : i + num_batches]

        # Read batch of files with optional parallelization
        if use_threads and len(batch_paths) > 1:
            batch_tables = run_parallel(
                read_parquet_file,
                batch_paths,
                self=self,
                include_file_path=include_file_path,
                n_jobs=-1,
                backend="threading",
                verbose=verbose,
                **kwargs,
            )
        else:
            batch_tables = [
                read_parquet_file(
                    p, self=self, include_file_path=include_file_path, **kwargs
                )
                for p in batch_paths
            ]

        if concat and len(batch_tables) > 1:
            yield pa.concat_tables(batch_tables, promote_options="permissive")
        else:
            yield batch_tables


def read_parquet(
    self,
    path: str | list[str],
    num_batches: int | None = None,
    include_file_path: bool = False,
    concat: bool = True,
    use_threads: bool = True,
    verbose: bool = False,
    **kwargs,
) -> pa.Table | list[pa.Table] | Generator[pa.Table | list[pa.Table], None, None]:
    """
    Read a Parquet file or a list of Parquet files. Optionally read in batches,
    returning a generator that sequentially yields data for specified number of files.

    Args:
        path: (str | list[str]) Path to the Parquet file(s).
        num_batches: (int, optional) Number of files to process in each batch. Defaults to 1.
        include_file_path: (bool, optional) If True, include 'file_path' column.
        concat: (bool, optional) If True, concatenate the batch Tables. Defaults to True.
        use_threads: (bool, optional) If True, use parallel processing within batches. Defaults to True.
        verbose: (bool, optional) If True, print verbose output. Defaults to False.
        **kwargs: Additional keyword arguments.

    Returns:
        pa.Table | list[pa.Table]:
            PyArrow Table or list of PyArrow Tables containing data from num_batches files.

    Yields:
        pa.Table | list[pa.Table]: PyArrow Table or list of PyArrow Tables containing data from num_batches files.
    """
    if num_batches is not None:
        yield from _read_parquet_batches(
            self=self,
            path=path,
            num_batches=num_batches,
            include_file_path=include_file_path,
            concat=concat,
            use_threads=use_threads,
            verbose=verbose,
            **kwargs,
        )
    return _read_parquet(
        self=self,
        path=path,
        include_file_path=include_file_path,
        concat=concat,
        use_threads=use_threads,
        verbose=verbose,
        **kwargs,
    )


def read_files(
    self,
    path: str | list[str],
    format: str,
    num_batches: int | None = None,
    include_file_path: bool = False,
    concat: bool = True,
    jsonlines: bool = False,
    use_threads: bool = True,
    verbose: bool = False,
    **kwargs,
) -> (
    pl.DataFrame
    | pa.Table
    | list[pl.DataFrame]
    | list[pa.Table]
    | Generator[
        pl.DataFrame | pa.Table | list[pa.Table] | list[pl.DataFrame], None, None
    ]
):
    """
    Read a file or a list of files of the given format.

    Args:
        path: (str | list[str]) Path to the file(s).
        format: (str) Format of the file.
        num_batches: (int, optional) Number of files to process in each batch. Defaults to None.
        include_file_path: (bool, optional) If True, return a DataFrame with a 'file_path' column.
            Defaults to False.
        concat: (bool, optional) If True, concatenate the DataFrames. Defaults to True.
        jsonlines: (bool, optional) If True, read JSON lines. Defaults to False.
        use_threads: (bool, optional) If True, read files in parallel. Defaults to True.
        verbose: (bool, optional) If True, print verbose output. Defaults to False.
        **kwargs: Additional keyword arguments.

    Returns:
        (pl.DataFrame | pa.Table | list[pl.DataFrame] | list[pa.Table]):
            Polars DataFrame, Pyarrow Table or list of DataFrames, LazyFrames or Tables.

    Yields:
        (pl.DataFrame | pa.Table):
            Polars DataFrame, Pyarrow Table or list of DataFrames, LazyFrames or Tables.
    """
    if format == "json":
        if num_batches is not None:
            yield from read_json(
                self=self,
                path=path,
                num_batches=num_batches,
                include_file_path=include_file_path,
                jsonlines=jsonlines,
                concat=concat,
                use_threads=use_threads,
                verbose=verbose,
                **kwargs,
            )
        return read_json(
            self,
            path,
            include_file_path=include_file_path,
            jsonlines=jsonlines,
            concat=concat,
            use_threads=use_threads,
            verbose=verbose,
            **kwargs,
        )
    elif format == "csv":
        if num_batches is not None:
            yield from read_csv(
                self=self,
                path=path,
                num_batches=num_batches,
                include_file_path=include_file_path,
                concat=concat,
                use_threads=use_threads,
                verbose=verbose,
                **kwargs,
            )
        return read_csv(
            self,
            path,
            include_file_path=include_file_path,
            use_threads=use_threads,
            concat=concat,
            verbose=verbose,
            **kwargs,
        )
    elif format == "parquet":
        if num_batches is not None:
            yield from read_parquet(
                self=self,
                path=path,
                num_batches=num_batches,
                include_file_path=include_file_path,
                concat=concat,
                use_threads=use_threads,
                verbose=verbose,
                **kwargs,
            )
        return read_parquet(
            self,
            path,
            include_file_path=include_file_path,
            use_threads=use_threads,
            concat=concat,
            verbose=verbose,
            **kwargs,
        )


def pyarrow_dataset(
    self,
    path: str,
    format="parquet",
    schema: pa.Schema | None = None,
    partitioning: str | list[str] | pds.Partitioning = None,
    **kwargs,
) -> pds.Dataset:
    """
    Create a pyarrow dataset.

    Args:
        path: (str) Path to the dataset.
        format: (str, optional) Format of the dataset. Defaults to 'parquet'.
        schema: (pa.Schema, optional) Schema of the dataset. Defaults to None.
        partitioning: (str | list[str] | pds.Partitioning, optional) Partitioning of the dataset.
            Defaults to None.
        **kwargs: Additional keyword arguments.

    Returns:
        (pds.Dataset): Pyarrow dataset.
    """
    return pds.dataset(
        path,
        filesystem=self,
        partitioning=partitioning,
        schema=schema,
        format=format,
        **kwargs,
    )


def pyarrow_parquet_dataset(
    self,
    path: str,
    schema: pa.Schema | None = None,
    partitioning: str | list[str] | pds.Partitioning = None,
    **kwargs,
) -> pds.Dataset:
    """
    Create a pyarrow dataset from a parquet_metadata file.

    Args:
        path: (str) Path to the dataset.
        schema: (pa.Schema, optional) Schema of the dataset. Defaults to None.
        partitioning: (str | list[str] | pds.Partitioning, optional) Partitioning of the dataset.
            Defaults to None.
        **kwargs: Additional keyword arguments.

    Returns:
        (pds.Dataset): Pyarrow dataset.
    """
    if not self.is_file(path):
        path = posixpath.join(path, "_metadata")
    return pds.dataset(
        path,
        filesystem=self,
        partitioning=partitioning,
        schema=schema,
        **kwargs,
    )


def write_parquet(
    self,
    data: pl.DataFrame | pl.LazyFrame | pa.Table | pd.DataFrame,
    path: str,
    schema: pa.Schema | None = None,
    **kwargs,
) -> pq.FileMetaData:
    """
    Write a DataFrame to a Parquet file.

    Args:
        data: (pl.DataFrame | pl.LazyFrame | pa.Table | pd.DataFrame) Data to write.
        path: (str) Path to write the data.
        schema: (pa.Schema, optional) Schema of the data. Defaults to None.
        **kwargs: Additional keyword arguments for `pq.write_table`.

    Returns:
        (pq.FileMetaData): Parquet file metadata.
    """
    if isinstance(data, pl.LazyFrame):
        data = data.collect()
    if isinstance(data, pl.DataFrame):
        data = data.to_arrow()
        data = data.cast(convert_large_types_to_standard(data.schema))
    elif isinstance(data, pd.DataFrame):
        data = pa.Table.from_pandas(data, preserve_index=False)

    if schema is not None:
        data = data.cast(schema)
    metadata = []
    pq.write_table(data, path, filesystem=self, metadata_collector=metadata, **kwargs)
    metadata = metadata[0]
    metadata.set_file_path(path)
    return metadata


def write_json(
    self,
    data: dict | pl.DataFrame | pl.LazyFrame | pa.Table | pd.DataFrame,
    path: str,
    append: bool = False,
) -> None:
    """
    Write a dictionary, DataFrame or Table to a JSON file.

    Args:
        data: (dict | pl.DataFrame | pl.LazyFrame | pa.Table | pd.DataFrame) Data to write.
        path: (str) Path to write the data.
        append: (bool, optional) If True, append to the file. Defaults to False.

    Returns:
        None
    """
    if isinstance(data, pl.LazyFrame):
        data = data.collect()
    if isinstance(data, pl.DataFrame):
        data = data.to_arrow()
        data = data.cast(convert_large_types_to_standard(data.schema)).to_pydict()
    elif isinstance(data, pd.DataFrame):
        data = pa.Table.from_pandas(data, preserve_index=False).to_pydict()
    elif isinstance(data, pa.Table):
        data = data.to_pydict()
    if append:
        with self.open(path, "ab") as f:
            f.write(orjson.dumps(data))
            f.write(b"\n")
    else:
        with self.open(path, "wb") as f:
            f.write(orjson.dumps(data))


def write_csv(
    self,
    data: pl.DataFrame | pl.LazyFrame | pa.Table | pd.DataFrame,
    path: str,
    **kwargs,
) -> None:
    """
    Write a DataFrame to a CSV file.

    Args:
        data: (pl.DataFrame | pl.LazyFrame | pa.Table | pd.DataFrame) Data to write.
        path: (str) Path to write the data.
        **kwargs: Additional keyword arguments for `pl.DataFrame.write_csv`.

    Returns:
        None
    """
    if isinstance(data, pl.LazyFrame):
        data = data.collect()
    if isinstance(data, pa.Table):
        data = pl.from_arrow(data)
    elif isinstance(data, pd.DataFrame):
        data = pl.from_pandas(data)

    with self.open(path, "w") as f:
        data.write_csv(f, **kwargs)


def write_file(
    self,
    data: pl.DataFrame | pl.LazyFrame | pa.Table | pd.DataFrame,
    path: str,
    format: str,
    **kwargs,
) -> None:
    """
    Write a DataFrame to a file in the given format.

    Args:
        data: (pl.DataFrame | pl.LazyFrame | pa.Table | pd.DataFrame) Data to write.
        path (str): Path to write the data.
        format (str): Format of the file.
        **kwargs: Additional keyword arguments.

    Returns:
        None
    """
    if format == "json":
        write_json(self, data, path, **kwargs)
    elif format == "csv":
        write_csv(self, data, path, **kwargs)
    elif format == "parquet":
        write_parquet(self, data, path, **kwargs)


def write_pyarrow_dataset(
    self,
    data: (
        pl.DataFrame
        | pa.Table
        | pa.RecordBatch
        | pa.RecordBatchReader
        | pd.DataFrame
        | list[pl.DataFrame]
        | list[pa.Table]
        | list[pa.RecordBatch]
        | list[pa.RecordBatchReader]
        | list[pd.DataFrame]
    ),
    path: str,
    basename: str | None = None,
    schema: pa.Schema | None = None,
    partition_by: str | list[str] | pds.Partitioning | None = None,
    partitioning_flavor: str = "hive",
    mode: str = "append",
    format: str | None = "parquet",
    compression: str = "zstd",
    **kwargs,
) -> list[pq.FileMetaData] | None:
    """
    Write a tabluar data to a PyArrow dataset.

    Args:
        data: (pl.DataFrame | pa.Table | pa.RecordBatch | pa.RecordBatchReader |
            pd.DataFrame | list[pl.DataFrame] | list[pa.Table] | list[pa.RecordBatch] |
            list[pa.RecordBatchReader] | list[pd.DataFrame]) Data to write.
        path: (str) Path to write the data.
        basename: (str, optional) Basename of the files. Defaults to None.
        schema: (pa.Schema, optional) Schema of the data. Defaults to None.
        partition_by: (str | list[str] | pds.Partitioning, optional) Partitioning of the data.
            Defaults to None.
        partitioning_flavor: (str, optional) Partitioning flavor. Defaults to 'hive'.
        mode: (str, optional) Write mode. Defaults to 'append'.
        format: (str, optional) Format of the data. Defaults to 'parquet'.
        compression: (str, optional) Compression algorithm. Defaults to 'zstd'.
        **kwargs: Additional keyword arguments for `pds.write_dataset`.

    Returns:
        (list[pq.FileMetaData] | None): List of Parquet file metadata or None.
    """
    if not isinstance(data, list):
        data = [data]

    if isinstance(data[0], pl.DataFrame):
        data = [dd.to_arrow() for dd in data]
        data = [dd.cast(convert_large_types_to_standard(dd.schema)) for dd in data]

    elif isinstance(data[0], pd.DataFrame):
        data = [pa.Table.from_pandas(dd, preserve_index=False) for dd in data]

    if mode == "delete_matching":
        existing_data_behavior = "delete_matching"
    elif mode == "append":
        existing_data_behavior = "overwrite_or_ignore"
    elif mode == "overwrite":
        self.rm(path, recursive=True)
        existing_data_behavior = "overwrite_or_ignore"
    else:
        existing_data_behavior = mode

    if basename is None:
        basename_template = (
            "data-"
            f"{dt.datetime.now().strftime('%Y%m%d_%H%M%S%f')[:-3]}-{uuid.uuid4().hex[:16]}-{{i}}.parquet"
        )
    else:
        basename_template = f"{basename}-{{i}}.parquet"

    file_options = pds.ParquetFileFormat().make_write_options(compression=compression)

    create_dir: bool = (False,)

    if hasattr(self, "fs"):
        if "local" in self.fs.protocol:
            create_dir = True
    else:
        if "local" in self.protocol:
            create_dir = True

    if format == "parquet":
        metadata = []

        def file_visitor(written_file):
            file_metadata = written_file.metadata
            file_metadata.set_file_path(written_file.path)
            metadata.append(file_metadata)

    pds.write_dataset(
        data=data,
        base_dir=path,
        basename_template=basename_template,
        partitioning=partition_by,
        partitioning_flavor=partitioning_flavor,
        filesystem=self,
        existing_data_behavior=existing_data_behavior,
        schema=schema,
        format=format,
        create_dir=create_dir,
        file_options=file_options,
        file_visitor=file_visitor if format == "parquet" else None,
        **kwargs,
    )
    if format == "parquet":
        return metadata


# patchen
DirFileSystem.ls = dir_ls_p
MonitoredSimpleCacheFileSystem.ls = mscf_ls_p
AbstractFileSystem.read_json_file = read_json_file
AbstractFileSystem.read_json = read_json
AbstractFileSystem.read_csv_file = read_csv_file
AbstractFileSystem.read_csv = read_csv
AbstractFileSystem.read_parquet_file = read_parquet_file
AbstractFileSystem.read_parquet = read_parquet
AbstractFileSystem.read_files = read_files
AbstractFileSystem.pyarrow_dataset = pyarrow_dataset
AbstractFileSystem.pyarrow_parquet_dataset = pyarrow_parquet_dataset
AbstractFileSystem.write_parquet = write_parquet
AbstractFileSystem.write_json = write_json
AbstractFileSystem.write_csv = write_csv
AbstractFileSystem.write_pyarrow_dataset = write_pyarrow_dataset


def get_filesystem(
    path: str | None = None,
    storage_options: (
        AwsStorageOptions
        | GitHubStorageOptions
        | GitLabStorageOptions
        | GcsStorageOptions
        | AzureStorageOptions
        | dict[str, str]
        | None
    ) = None,
    dirfs: bool = True,
    cached: bool = False,
    cache_storage: str | None = None,
    **storage_options_kwargs,
) -> AbstractFileSystem:
    """
    Get a filesystem based on the given path.

    Args:
        path: (str, optional) Path to the filesystem. Defaults to None.
        storage_options: (AwsStorageOptions | GitHubStorageOptions | GitLabStorageOptions |
            GcsStorageOptions | AzureStorageOptions | dict[str, str], optional) Storage options.
            Defaults to None.
        dirfs: (bool, optional) If True, return a DirFileSystem. Defaults to True.
        cached: (bool, optional) If True, use a cached filesystem. Defaults to False.
        cache_storage: (str, optional) Path to the cache storage. Defaults to None.
        **storage_options_kwargs: Additional keyword arguments for the storage options.

    """
    pp = infer_storage_options(path)
    protocol = pp.get("protocol")
    path = pp.get("host", "") + pp.get("path", "")

    if protocol == "file" or protocol == "local":
        fs = filesystem(protocol)
        fs.is_cache_fs = False
        if dirfs:
            fs = DirFileSystem(path=path, fs=fs)
            fs.is_cache_fs = False
        return fs

    if storage_options is None:
        storage_options = get_storage_options(protocol, **storage_options_kwargs)

    fs = storage_options.to_filesystem()
    fs.is_cache_fs = False
    if dirfs:
        fs = DirFileSystem(path=path, fs=fs)
        fs.is_cache_fs = False
    if cached:
        if cache_storage is None:
            cache_storage = (Path.cwd() / path).as_posix()
        fs = MonitoredSimpleCacheFileSystem(fs=fs, cache_storage=cache_storage)
        fs.is_cache_fs = True

    return fs



from sanic import Sanic

from ..scheduler import SchedulerManager
from .api.cfg import bp as bp_api_cfg
from .api.pipeline import bp as bp_api_pipeline
from .api.scheduler import bp as bp_api_scheduler


def setup(
    app: Sanic,
    base_dir: str | None = None,
    storage_options: dict | None = None,
    #cfg_dir: str = "conf",
    #pipelines_dir: str = "pipelines",
) -> Sanic:

    app.config.BASE_DIR = base_dir
    app.config.STORAGE_OPTIONS = storage_options
    #app.config.CFG_DIR = cfg_dir
    #app.config.PIPELINES_DIR = pipelines_dir

    @app.listener("before_server_start")
    def init_scheduler(app, loop):

        app.ctx.scheduler = SchedulerManager(
            base_dir=base_dir, storage_options=storage_options
        )
        app.ctx.scheduler.start_worker(background=True)

    @app.listener("before_server_stop")
    def cleanup_scheduler(app, loop):
        if hasattr(app.ctx, "scheduler"):
            app.ctx.scheduler.stop_worker()

    app.blueprint(bp_api_pipeline)
    app.blueprint(bp_api_scheduler)
    app.blueprint(bp_api_cfg)

    return app



import typer

from ..cli.utils import parse_dict_or_list_param
from ..cfg import Config

app = typer.Typer(help="Config management commands")


# @app.command()
# def get_project(request) -> json:
#     cfg = request.app.ctx.pipeline_manager.cfg.project.to_dict()
#     # cfg.pop("fs")
#     return json({"cfg": cfg})


# @bp.get("/pipeline/<pipeline_name>")
# async def get_pipeline(request, pipeline_name) -> json:
#     if pipeline_name != request.app.ctx.pipeline_manager.cfg.pipeline.name:
#         request.app.ctx.pipeline_manager.load_config(pipeline_name)
#     cfg = request.app.ctx.pipeline_manager.cfg.pipeline.to_dict()
#     return json({"cfg": cfg})


# @bp.post("/pipeline/<pipeline_name>")
# @openapi.body({"application/json": PipelineConfig}, required=True)
# @validate(json=PipelineConfig)
# async def update_pipeline(request, pipeline_name, body: PipelineConfig) -> json:
#     data = request.json
#     if pipeline_name != request.app.ctx.pipeline_manager.cfg.pipeline.name:
#         request.app.ctx.pipeline_manager.load_config(pipeline_name)
#     cfg = request.app.ctx.pipeline_manager.cfg.pipeline.copy()
#     cfg.update(data)
#     try:
#         cfg.to_yaml(
#             posixpath.join(
#                 "pipelines",
#                 pipeline_name + ".yml",
#             ),
#             fs=request.app.ctx.pipeline_manager.cfg.fs,
#         )
#     except NotImplementedError as e:
#         raise SanicException(f"Update failed. {e}", status_code=404)
#     cfg
#     return json({"cfg": cfg})



from pathlib import Path

from pydala.catalog import Catalog


def load_catalog(namespace: str | None = None, path: str | None = None) -> Catalog:
    """
    Load a catalog from a YAML file.

    Args:
        namespace (str | None, optional): The namespace of the catalog. Defaults to None.
        path (str | None, optional): The path to the YAML file. If not provided, the function will search
            for a file named "catalog*.y*ml" in the parent directories. Defaults to None.

    Returns:
        Catalog: The loaded catalog object.
    """
    if path is None:
        path = list(Path.cwd().rglob("catalog*.y*ml"))
        if not len(path):
            return
        path = path[0]

    return Catalog(path=path, namespace=namespace)


CATALOG = load_catalog()



import ast
import json
import re

from loguru import logger


# Parse additional parameters
def parse_param_dict(param_str: str | None) -> dict:
    """Helper to parse parameter dictionaries"""
    if not param_str:
        return {}
    return dict(param.split("=") for param in param_str.split(","))


def parse_dict_or_list_param(
    value: str | None = None, param_type: str = "dict"
) -> list | dict | None:
    """
    Parse dictionary or list parameters from various input formats.

    Supports:
    - JSON string
    - Python literal (dict/list)
    - Comma-separated key=value pairs (for dicts)
    - Comma-separated values (for lists)
    - List-like string with or without quotes

    Args:
        value (str, optional): Input string to parse
        param_type (str): Type of parameter to parse ('dict' or 'list')

    Returns:
        dict | list | None: Parsed parameter or None if parsing fails
    """

    def convert_string_booleans(obj):
        if isinstance(obj, dict):
            return {k: convert_string_booleans(v) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [convert_string_booleans(item) for item in obj]
        elif isinstance(obj, str):
            if obj.lower() == "true":
                return True
            elif obj.lower() == "false":
                return False
        return obj

    if value is None:
        return None

    try:
        # Try parsing as JSON first
        parsed = json.loads(value)
        return convert_string_booleans(parsed)
    except json.JSONDecodeError:
        try:
            # Try parsing as Python literal
            parsed = ast.literal_eval(value)

            # Validate type
            if param_type == "dict" and not isinstance(parsed, dict):
                raise ValueError(f"Expected dict, got {type(parsed)}")
            elif param_type == "list" and not isinstance(parsed, list):
                raise ValueError(f"Expected list, got {type(parsed)}")

            return convert_string_booleans(parsed)
        except (SyntaxError, ValueError):
            # For dicts, try parsing as comma-separated key=value pairs
            if param_type == "dict" and "=" in value:
                parsed = dict(
                    pair.split("=", 1) for pair in value.split(",") if pair.strip()
                )
                return convert_string_booleans(parsed)

            # For lists, try multiple parsing strategies
            if param_type == "list":
                # Remove surrounding square brackets and whitespace
                value = value.strip()
                if value.startswith("[") and value.endswith("]"):
                    value = value[1:-1].strip()

                # Parse list-like string with or without quotes
                # This regex handles: a,b | 'a','b' | "a","b" | a, b | 'a', 'b'
                list_items = re.findall(r"['\"]?(.*?)['\"]?(?=\s*,|\s*$)", value)

                # Remove any empty strings and strip whitespace
                parsed = [item.strip() for item in list_items if item.strip()]

                return convert_string_booleans(parsed)

            # If all parsing fails, log warning and return None
            logger.warning(f"Could not parse {param_type} parameter: {value}")
            return None



from pydantic import Field

from ..base import BaseConfig


class ProjectTrackerConfig(BaseConfig):
    username: str | None = Field(default=None)
    api_url: str = "http://localhost:8241"
    ui_url: str = "http://localhost:8242"
    api_key: str | None = Field(default=None)
    verify: bool = False



import importlib.util
import sys

if importlib.util.find_spec("apscheduler"):
    from apscheduler import Scheduler, current_scheduler
    from apscheduler.executors.async_ import AsyncJobExecutor
    from apscheduler.executors.subprocess import ProcessPoolJobExecutor
    from apscheduler.executors.thread import ThreadPoolJobExecutor

    from .utils.monkey import patch_pickle

    patch_pickle()

else:
    raise ImportError(
        "APScheduler is not installed. Please install it using `pip install"
        "'apscheduler>4.0.0a1'`, 'conda install apscheduler4' or `pip install flowerpower[scheduler]`"
    )

import posixpath
import uuid
from pathlib import Path
from typing import Any

from fsspec.spec import AbstractFileSystem
from loguru import logger

from .cfg import Config
from .utils.datastore import setup_data_store
from .utils.eventbroker import setup_event_broker
from .filesystem import get_filesystem
from .utils.scheduler import display_jobs, display_schedules, display_tasks


class SchedulerManager(Scheduler):
    def __init__(
        self,
        name: str | None = None,
        base_dir: str | None = None,
        storage_options: dict = {},
        fs: AbstractFileSystem | None = None,
        **kwargs,
    ):
        """
        Initializes the Scheduler object.

        Args:
            name (str | None, optional): The name of the scheduler. Defaults to None.
            base_dir (str | None, optional): The flowerpower base path. Defaults to None.
            **kwargs: Additional keyword arguments.

        """
        self.name = name or ""
        self._base_dir = base_dir or str(Path.cwd())
        self._storage_options = storage_options
        if fs is None:
            fs = get_filesystem(self._base_dir, **self._storage_options)
        self._fs = fs

        self._conf_path = "conf"
        self._pipelines_path = "pipelines"  # or pipelines_path

        self._sync_fs()
        self.load_config()

        self._setup_data_store()
        self._setup_event_broker()
        self._setup_job_executors()

        super().__init__(
            data_store=self._data_store,
            event_broker=self._event_broker,
            job_executors=self._job_executors,
            identity=self.name,
            logger=logger,
            cleanup_interval=self.cfg.project.worker.cleanup_interval,
            max_concurrent_jobs=self.cfg.project.worker.max_concurrent_jobs,
            **kwargs,
        )

        sys.path.append(self._pipelines_path)

    def _sync_fs(self):
        """
        Sync the filesystem.

        Returns:
            None
        """
        if self._fs.is_cache_fs:
            self._fs.sync()

        modules_path = posixpath.join(self._fs.path, self._pipelines_path)
        if modules_path not in sys.path:
            sys.path.append(modules_path)

    def load_config(self):
        """
        Load the configuration file.

        This method loads the configuration file specified by the `_conf_dir` attribute and
        assigns it to the `cfg` attribute.

        """
        self.cfg = Config.load(base_dir=self._base_dir, fs=self._fs)

    def _setup_data_store(self):
        """
        Sets up the data store based on the configuration.

        If the configuration specifies a data store type, it retrieves the data store and SQLA engine
        using the `get_data_store` function. The data store type is obtained from the `type` key in the
        `data_store` section of the configuration. The engine or URI is obtained from the `uri` key in
        the `data_store` section of the configuration.

        Returns:
            None
        """

        self._data_store, self._sqla_engine = setup_data_store(
            type=self.cfg.project.worker.data_store.get("type", "memory"),
            engine_or_uri=self.cfg.project.worker.data_store.get("uri", None),
        )

    def _setup_event_broker(self):
        """
        Sets up the event broker based on the configuration settings.

        If the 'event_broker' key is present in the configuration, it checks for the 'type' key.
        If the 'type' key is present, it initializes the event broker using the specified type and other optional
            parameters.
        The optional parameters include 'uri', 'sqla_engine', 'host', 'port', 'username', and 'password'.

        Parameters:
            None

        Returns:
            None
        """
        self._event_broker = setup_event_broker(
            type=self.cfg.project.worker.event_broker.get("type", "memory"),
            uri=self.cfg.project.worker.event_broker.get("uri", None),
            sqla_engine=self._sqla_engine,
            host=self.cfg.project.worker.event_broker.get("host", None),
            port=self.cfg.project.worker.event_broker.get("port", 0),
            username=self.cfg.project.worker.event_broker.get("username", None),
            password=self.cfg.project.worker.event_broker.get("password", None),
        )

    def _setup_job_executors(self):
        self._job_executors = {
            "async": AsyncJobExecutor(),
            "threadpool": ThreadPoolJobExecutor(),
            "processpool": ProcessPoolJobExecutor(),
        }

    def start_worker(
        self,
        background: bool = False,
    ):
        """
        Starts the worker.

        Parameters:
            background (bool): If True, starts the worker in the background. If False, runs the worker until stopped.
        """
        if background:
            self.start_in_background()
        else:
            self.run_until_stopped()

    def stop_worker(self) -> None:
        """
        Stops the worker and closes the exit stack.

        Returns:
            None
        """
        self.stop()
        self._exit_stack.close()

    def remove_all_schedules(self):
        """
        Removes all schedules from the scheduler.

        This method iterates over all schedules in the scheduler and removes them one by one.

        Parameters:
            None

        Returns:
            None
        """
        for sched in self.get_schedules():
            self.remove_schedule(sched.id)

    def get_tasks(self, as_dict: bool = False):
        """
        Get all tasks from the scheduler.

        Args:
            as_dict (bool, optional): Whether to return the tasks as a list of dictionaries. Defaults to False.

        Returns:
            list: A list of tasks.
        """
        tasks = super().get_tasks()
        if as_dict:
            return [task.to_dict() for task in tasks]
        return super().get_tasks()

    def get_schedules(self, as_dict: bool = False):
        """
        Get all schedules from the scheduler.

        Args:
            as_dict (bool, optional): Whether to return the schedules as a list of dictionaries. Defaults to False.

        Returns:
            list: A list of schedules.
        """
        schedules = super().get_schedules()
        if as_dict:
            return [schedule.to_dict() for schedule in schedules]
        return schedules

    def get_schedule(self, id: str, as_dict: bool = False):
        """
        Get a schedule by ID.

        Args:
            id (str): The ID of the schedule.
            as_dict (bool, optional): Whether to return the schedule as a dictionary. Defaults to False.

        Returns:
            Schedule: The schedule object.
        """
        schedule = super().get_schedule(id)
        if as_dict:
            return schedule.to_dict()
        return schedule

    def get_jobs(self, as_dict: bool = False):
        """
        Get all jobs from the scheduler.

        Args:
            as_dict (bool, optional): Whether to return the jobs as a list of dictionaries. Defaults to False.

        Returns:
            list: A list of jobs.
        """
        jobs = super().get_jobs()
        if as_dict:
            return [job.to_dict() for job in jobs]
        return jobs

    def show_schedules(self):
        """
        Shows all schedules in the scheduler.

        This method iterates over all schedules in the scheduler and prints their details.

        Parameters:
            None
        """
        display_schedules(self.get_schedules())

    def show_tasks(self):
        """
        Shows all tasks in the scheduler.

        This method iterates over all tasks in the scheduler and prints their details.

        Parameters:
            None
        """
        display_tasks(self.get_tasks())

    def show_jobs(self):
        """
        Shows all jobs in the scheduler.

        This method iterates over all jobs in the scheduler and prints their details.

        Parameters:
            None
        """
        display_jobs(self.get_jobs())


# Wrapper functions for backward compatibility
def get_schedule_manager(
    name: str | None = None,
    base_dir: str | None = None,
    storage_options: dict = {},
    fs: AbstractFileSystem | None = None,
    **kwargs,
) -> SchedulerManager:
    """
    Returns a SchedulerManager instance with the specified name and base path.

    Args:
        name (str | None, optional): The name of the scheduler manager. Defaults to None.
        base_dir (str | None, optional): The base path for the scheduler manager. Defaults to None.
        storage_options (dict, optional): The storage options. Defaults to {}.
        fs (AbstractFileSystem, optional): The filesystem. Defaults to None.
        **kwargs: Additional keyword arguments to be passed to the SchedulerManager constructor.

    Returns:
        SchedulerManager: The initialized SchedulerManager instance.
    """
    manager = SchedulerManager(
        name, base_dir, storage_options=storage_options, fs=fs, **kwargs
    )
    return manager


def get_current_scheduler_manager() -> SchedulerManager | None:
    """
    Returns the current scheduler manager.

    Returns:
        The current scheduler manager if available, otherwise None.
    """
    return current_scheduler.get()


def start_worker(
    name: str | None = None,
    base_dir: str | None = None,
    background: bool = False,
    storage_options: dict = {},
    fs: AbstractFileSystem | None = None,
    **kwargs,
) -> SchedulerManager:
    """
    Start the scheduler.

    Args:
        name (str | None, optional): The name of the scheduler. Defaults to None.
        base_dir (str | None, optional): The base path for the scheduler. Defaults to None.
        background (bool, optional): Whether to start the scheduler in the background. Defaults to False.
        storage_options (dict, optional): The storage options. Defaults to {}.
        fs (AbstractFileSystem, optional): The filesystem. Defaults to None.
        **kwargs: Additional keyword arguments.

    Returns:
        SchedulerManager: The scheduler instance.
    """
    # manager = get_schedule_manager(name, base_dir, role="worker", storage_options=storage_options, fs=fs, **kwargs)
    with SchedulerManager(
        name, base_dir, storage_options=storage_options, fs=fs, **kwargs
    ) as manager:
        manager.start_worker(background=background)


def remove_all_schedules(
    name: str | None = None,
    base_dir: str | None = None,
    storage_options: dict = {},
    fs: AbstractFileSystem | None = None,
    **kwargs,
):
    """
    Remove all schedules.

    Args:
        name (str | None, optional): The name of the scheduler. Defaults to None.
        base_dir (str | None, optional): The base path. Defaults to None.
        storage_options (dict, optional): The storage options. Defaults to {}.
        fs (AbstractFileSystem, optional): The filesystem. Defaults to None.
        **kwargs: Additional keyword arguments.
    """
    with SchedulerManager(
        name,
        base_dir,
        role="scheduler",
        storage_options=storage_options,
        fs=fs,
        **kwargs,
    ) as manager:
        manager.remove_all_schedules()


def add_schedule(
    name: str | None = None,
    base_dir: str | None = None,
    storage_options: dict = {},
    fs: AbstractFileSystem | None = None,
    **kwargs,
) -> str:
    """
    Add a schedule to the scheduler.

    Args:
        name (str, optional): The name of the schedule. Defaults to None.
        base_dir (str, optional): The base path for the schedule. Defaults to None.
        storage_options (dict, optional): The storage options. Defaults to {}.
        fs (AbstractFileSystem, optional): The filesystem. Defaults to None.
        **kwargs: Arbitrary keyword arguments.

    Returns:
        str: The ID of the added schedule.
    """
    with SchedulerManager(name, base_dir, role="scheduler") as manager:
        id_ = manager.add_schedule(storage_options=storage_options, fs=fs, **kwargs)
    return id_


def add_job(
    name: str | None = None,
    base_dir: str | None = None,
    storage_options: dict = {},
    fs: AbstractFileSystem | None = None,
    **kwargs,
) -> uuid.UUID:
    """
    Add a job to the scheduler. Executes the job immediatly and returns the job id (UUID).
    The job result will be stored in the data store for the given `result_expiration_time` and
    can be fetched using the job id (UUID).

    Args:
        name (str | None): The name of the job. Defaults to None.
        base_dir (str | None): The base path for the job. Defaults to None.
        storage_options (dict, optional): The storage options. Defaults to {}.
        fs (AbstractFileSystem, optional): The filesystem. Defaults to None.
        **kwargs: Arbitrary keyword arguments.

    Returns:
        uuid.UUID: The ID of the added job.
    """
    with SchedulerManager(name, base_dir, role="scheduler") as manager:
        id_ = manager.add_job(storage_options=storage_options, fs=fs, **kwargs)
    return id_


def run_job(
    name: str | None = None,
    base_dir: str | None = None,
    storage_options: dict = {},
    fs: AbstractFileSystem | None = None,
    **kwargs,
) -> Any:
    """
    Run a job using the SchedulerManager. Executes the job immediatly and returns the result of the job.

    Args:
        name (str, optional): The name of the job. Defaults to None.
        base_dir (str, optional): The base path of the job. Defaults to None.
        storage_options (dict, optional): The storage options. Defaults to {}.
        fs (AbstractFileSystem, optional): The filesystem. Defaults to None.
        **kwargs: Arbitrary keyword arguments.

    Returns:
        Any: The SchedulerManager instance.

    """
    with SchedulerManager(name, base_dir, role="scheduler") as manager:
        result = manager.run_job(storage_options=storage_options, fs=fs, **kwargs)

    return result


def get_schedules(
    name: str | None = None,
    base_dir: str | None = None,
    storage_options: dict = {},
    fs: AbstractFileSystem | None = None,
    **kwargs,
):
    """
    Get all schedules from the scheduler.

    Args:
        name (str | None, optional): The name of the scheduler. Defaults to None.
        base_dir (str | None, optional): The base path. Defaults to None.
        storage_options (dict, optional): The storage options. Defaults to {}.
        fs (AbstractFileSystem, optional): The filesystem. Defaults to None.
        **kwargs: Additional keyword arguments.

    Returns:
        list: A list of schedules.
    """
    with SchedulerManager(
        name,
        base_dir,
        role="scheduler",
        storage_options=storage_options,
        fs=fs,
        **kwargs,
    ) as manager:
        schedules = manager.get_schedules()
    return schedules


def get_tasks(
    name: str | None = None,
    base_dir: str | None = None,
    storage_options: dict = {},
    fs: AbstractFileSystem | None = None,
    **kwargs,
):
    """
    Get all tasks from the scheduler.

    Args:
        name (str | None, optional): The name of the scheduler. Defaults to None.
        base_dir (str | None, optional): The base path. Defaults to None.
        storage_options (dict, optional): The storage options. Defaults to {}.
        fs (AbstractFileSystem, optional): The filesystem. Defaults to None.
        **kwargs: Additional keyword arguments.

    Returns:
        list: A list of tasks.
    """
    with SchedulerManager(
        name,
        base_dir,
        role="scheduler",
        storage_options=storage_options,
        fs=fs,
        **kwargs,
    ) as manager:
        tasks = manager.get_tasks()
    return tasks


def get_jobs(
    name: str | None = None,
    base_dir: str | None = None,
    storage_options: dict = {},
    fs: AbstractFileSystem | None = None,
    **kwargs,
):
    """
    Get all jobs from the scheduler.

    Args:
        name (str | None, optional): The name of the scheduler. Defaults to None.
        base_dir (str | None, optional): The base path. Defaults to None.
        storage_options (dict, optional): The storage options. Defaults to {}.
        fs (AbstractFileSystem, optional): The filesystem. Defaults to None.
        **kwargs: Additional keyword arguments.

    Returns:
        list: A list of jobs.
    """
    with SchedulerManager(
        name,
        base_dir,
        role="scheduler",
        storage_options=storage_options,
        fs=fs,
        **kwargs,
    ) as manager:
        jobs = manager.get_jobs()
    return jobs


def get_job_result(
    job_id: str,
    name: str | None = None,
    base_dir: str | None = None,
    storage_options: dict = {},
    fs: AbstractFileSystem | None = None,
    **kwargs,
) -> Any:
    """
    Get the result of a job using the job id.

    Args:
        job_id (str): The ID of the job.
        name (str | None, optional): The name of the job. Defaults to None.
        base_dir (str | None, optional): The base path. Defaults to None.
        storage_options (dict, optional): The storage options. Defaults to {}.
        fs (AbstractFileSystem, optional): The filesystem. Defaults to None.
        **kwargs: Additional keyword arguments.

    Returns:
        Any: The result of the job.
    """
    with SchedulerManager(
        name,
        base_dir,
        role="scheduler",
        storage_options=storage_options,
        fs=fs,
        **kwargs,
    ) as manager:
        result = manager.get_job_result(job_id)
    return result


def show_schedules(
    name: str | None = None,
    base_dir: str | None = None,
    storage_options: dict = {},
    fs: AbstractFileSystem | None = None,
    **kwargs,
):
    """
    Show all schedules in the scheduler.

    Args:
        name (str | None, optional): The name of the scheduler. Defaults to None.
        base_dir (str | None, optional): The base path. Defaults to None.
        storage_options (dict, optional): The storage options. Defaults to {}.
        fs (AbstractFileSystem, optional): The filesystem. Defaults
        **kwargs: Additional keyword arguments.
    """
    with SchedulerManager(
        name,
        base_dir,
        role="scheduler",
        storage_options=storage_options,
        fs=fs,
        **kwargs,
    ) as manager:
        manager.show_schedules()


def show_tasks(
    name: str | None = None,
    base_dir: str | None = None,
    storage_options: dict = {},
    fs: AbstractFileSystem | None = None,
    **kwargs,
):
    """
    Show all tasks in the scheduler.

    Args:
        name (str | None, optional): The name of the scheduler. Defaults to None.
        base_dir (str | None, optional): The base path. Defaults to None.
        storage_options (dict, optional): The storage options. Defaults to {}.
        fs (AbstractFileSystem, optional): The filesystem. Defaults
        **kwargs: Additional keyword arguments.
    """
    with SchedulerManager(
        name,
        base_dir,
        role="scheduler",
        storage_options=storage_options,
        fs=fs,
        **kwargs,
    ) as manager:
        manager.show_tasks()


def show_jobs(
    name: str | None = None,
    base_dir: str | None = None,
    storage_options: dict = {},
    fs: AbstractFileSystem | None = None,
    **kwargs,
):
    """
    Show all jobs in the scheduler.

    Args:
        name (str | None, optional): The name of the scheduler. Defaults to None.
        base_dir (str | None, optional): The base path. Defaults to None.
        storage_options (dict, optional): The storage options. Defaults to {}.
        fs (AbstractFileSystem, optional): The filesystem. Defaults
        **kwargs: Additional keyword arguments.
    """
    with SchedulerManager(
        name,
        base_dir,
        role="scheduler",
        storage_options=storage_options,
        fs=fs,
        **kwargs,
    ) as manager:
        manager.show_jobs()



# import datetime as dt
import os

import duckdb
import pandas as pd
import polars as pl
import pyarrow as pa
import pyarrow.dataset as pds

# from hamilton.function_modifiers import dataloader
from pydantic import BaseModel


class DuckDBLoader(BaseModel):
    path: str | None = None
    read_only: bool = False
    conn: duckdb.DuckDBPyConnection | None = None
    host: str | None = None
    port: int | None = None
    socket: str | None = None
    user: str | None = None
    password: str | None = None
    database: str | None = None

    def _is_sqlite_file(self, path: str) -> bool:
        """Check if file is SQLite by reading first 16 bytes for SQLite header."""
        if not os.path.exists(path):
            return False
        with open(path, "rb") as f:
            header = f.read(16)
        return header.startswith(b"SQLite format 3\x00")

    def model_post_init(self, __context):
        if self.path is None:
            self.path = ":memory:"
            self.conn = duckdb.connect(self.path, self.read_only)
        else:
            if os.path.exists(self.path):
                if self._is_sqlite_file(self.path):
                    # For SQLite files, create in-memory DuckDB and attach SQLite
                    self.conn = duckdb.connect(":memory:", self.read_only)
                    self.conn.execute("INSTALL sqlite; LOAD sqlite;")
                    self.conn.execute(
                        f"ATTACH '{self.path}' AS sqlite_db (TYPE SQLITE, READ_ONLY);"
                    )
                else:
                    # For DuckDB files, connect directly
                    self.conn = duckdb.connect(self.path, self.read_only)
            else:
                # For new files, create new DuckDB database
                self.conn = duckdb.connect(self.path, self.read_only)

        if self.host is not None:
            self._update_from_env()

        if self.host is not None:
            if self.port == 5432:
                self._attach_postgres()
            elif self.port == 3306:
                self._attach_mysql()

    def _update_from_env(self):
        self.host = (
            os.getenv("POSTGRES_HOST") or os.getenv("PGHOST") or os.getenv("MYSQL_HOST")
        )
        self.port = (
            os.getenv("POSTGRES_PORT")
            or os.getenv("PGPORT")
            or os.getenv("MYSQL_TCP_PORT")
        )
        self.user = (
            os.getenv("POSTGRES_USER") or os.getenv("PGUSER") or os.getenv("MYSQL_USER")
        )
        self.password = (
            os.getenv("POSTGRES_PASSWORD")
            or os.getenv("PGPASSWORD")
            or os.getenv("MYSQL_PW")
        )
        self.database = (
            os.getenv("POSTGRES_DB")
            or os.getenv("PGDATABASE")
            or os.getenv("MYSQL_DATABASE")
        )
        self.socket = os.getenv("MYSQL_UNIX_PORT")

    def _attach_postgres(
        self,
    ):
        self.conn.execute("INSTALL postgres; LOAD postgres;")
        connection_string = ""
        for k, v in {
            "host": self.host,
            "port": self.port,
            "user": self.user,
            "password": self.password,
            "database": self.database,
        }.items():
            if v is not None:
                connection_string += f"{k}={v} "

        self.conn.execute(
            f"ATTACH {connection_string} AS postgres_db (TYPE POSTGRES, READ_ONLY);"
        )

    def _attach_mysql(
        self,
    ):
        self.conn.execute("INSTALL mysql; LOAD mysql;")
        connection_string = ""
        for k, v in {
            "host": self.host,
            "port": self.port,
            "socket": self.socket,
            "user": self.user,
            "password": self.password,
            "database": self.database,
        }.items():
            if v is not None:
                connection_string += f"{k}={v} "
        self.conn.execute(
            f"ATTACH {connection_string} AS mysql_db (TYPE MYSQL, READ_ONLY);"
        )

    def to_duckdb_relation(self, query: str) -> duckdb.DuckDBPyRelation:
        return self.conn.sql(query)

    def to_pandas(self, query: str) -> pd.DataFrame:
        return self.conn.sql(query).df()

    def to_polars(self, query: str) -> pl.DataFrame:
        return self.conn.sql(query).pl()

    def to_pyarrow_table(self, query: str) -> pa.Table:
        return self.conn.sql(query).arrow()

    def to_table(self, query: str) -> pa.Table:
        return self.conn.sql(query).arrow()

    def to_pyarrow_dataset(self, query: str) -> pds.Dataset:
        return pds.dataset(self.conn.sql(query).arrow())

    def to_dataset(self, query: str) -> pds.Dataset:
        return pds.dataset(self.conn.sql(query).arrow())

    def to_record_batch_reader(
        self, query: str, batch_size: int | None = None
    ) -> pa.RecordBatchReader:
        return self.conn.sql(query).record_batch(batch_size=batch_size)


# @dataloader()
# def load_from_duckdb(
#     query: str,
#     path: str | None = None,
#     conn: duckdb.DuckDBPyConnection | None = None,
#     **kwargs,
# ) -> tuple[duckdb.DuckDBPyRelation, dict]:
#     """
#     Load data from a DuckDB database.

#     Args:
#         query: (str) SQL query.
#         path: (str) Path to the database file. If None, an in-memory database is created.
#         conn: (DuckDBPyConnection) DuckDB connection.
#         **kwargs: Additional keyword arguments to pass to `duckdb.connect`.

#     Returns:
#         table: (DuckDBPyRelation) DuckDB relation.
#         metadata: (dict) Metadata dictionary.
#     """
#     if conn is None:
#         conn = duckdb.connect(path, **kwargs)

#     table = conn.sql(query)
#     metadata = {
#         "path": path,
#         "format": "duckdb",
#         "timestamp": dt.datetime.now().timestamp(),
#         "query": query,
#     }
#     return table, metadata


# @dataloader()
# def load_from_sqlite(
#     query: str,
#     path: str | None = None,
#     conn: duckdb.DuckDBPyConnection | None = None,
#     **kwargs,
# ) -> tuple[duckdb.DuckDBPyRelation, dict]:
#     """
#     Load data from a SQLite database.

#     Args:
#         query: (str) SQL query.
#         path: (str) Path to the database file. If None, an in-memory database is created.
#         conn: (DuckDBPyConnection) DuckDB connection.
#         **kwargs: Additional keyword arguments to pass to `duckdb.connect`.

#     Returns:
#         table: (DuckDBPyRelation) DuckDB relation.
#         metadata: (dict) Metadata dictionary.
#     """
#     if conn is None:
#         conn = duckdb.connect(path, **kwargs)

#     table = conn.sql(query)
#     metadata = {
#         "path": path,
#         "format": "sqlite",
#         "timestamp": dt.datetime.now().timestamp(),
#         "query": query,
#     }
#     return table, metadata


# @dataloader()
# def load_from_postgres(
#     query: str,
#     path: str | None = None,
#     conn: duckdb.DuckDBPyConnection | None = None,
#     host: str | None = None,
#     port: int | None = None,
#     user: str | None = None,
#     password: str | None = None,
#     database: str | None = None,
#     **kwargs,
# ) -> tuple[duckdb.DuckDBPyRelation, dict]:
#     """
#     Load data from a PostgreSQL database.

#     Args:
#         query: (str) SQL query.
#         path: (str) Path to the database file. If None, an in-memory database is created.
#         conn: (DuckDBPyConnection) DuckDB connection.
#         host: (str) PostgreSQL host.
#         port: (int) PostgreSQL port.
#         user: (str) PostgreSQL user.
#         password: (str) PostgreSQL password.
#         database: (str) PostgreSQL database.
#         **kwargs: Additional keyword arguments to pass to `duckdb.connect`.

#     Returns:
#         table: (DuckDBPyRelation) DuckDB relation.
#         metadata: (dict) Metadata dictionary.
#     """
#     if conn is None:
#         if path is None:
#             path = ":memory:"

#         conn = duckdb.connect(path, **kwargs)

#     conn.execute("INSTALL postgres; LOAD postgres;")
#     connection_string = " ".join(
#         f"{k}={v}"
#         for k, v in {
#             "host": host,
#             "port": port,
#             "user": user,
#             "password": password,
#             "database": database,
#         }.items()
#         if v is not None
#     )
#     conn.execute(
#         f"ATTACH {connection_string} AS postgres_db (TYPE POSTGRES, READ_ONLY);"
#     )

#     table = conn.sql(query)
#     metadata = {
#         "path": path,
#         "format": "postgres",
#         "timestamp": dt.datetime.now().timestamp(),
#         "query": query,
#     }
#     return table, metadata


# @dataloader()
# def load_from_mysql(
#     query: str,
#     path: str | None = None,
#     conn: duckdb.DuckDBPyConnection | None = None,
#     host: str | None = None,
#     port: int | None = None,
#     user: str | None = None,
#     password: str | None = None,
#     database: str | None = None,
#     **kwargs,
# ) -> tuple[duckdb.DuckDBPyRelation, dict]:
#     """
#     Load data from a MySQL database.

#     Args:
#         query: (str) SQL query.
#         path: (str) Path to the database file. If None, an in-memory database is created.
#         conn: (DuckDBPyConnection) DuckDB connection.
#         host: (str) MySQL host.
#         port: (int) MySQL port.
#         user: (str) MySQL user.
#         password: (str) MySQL password.
#         database: (str) MySQL database.
#         **kwargs: Additional keyword arguments to pass to `duckdb.connect`.

#     """
#     if conn is None:
#         if path is None:
#             path = ":memory:"

#         conn = duckdb.connect(path, **kwargs)

#     conn.execute("INSTALL mysql; LOAD mysql;")
#     connection_string = " ".join(
#         f"{k}={v}"
#         for k, v in {
#             "host": host,
#             "port": port,
#             "user": user,
#             "password": password,
#             "database": database,
#         }.items()
#         if v is not None
#     )
#     conn.execute(f"ATTACH {connection_string} AS mysql_db (TYPE MYSQL, READ_ONLY);")

#     table = conn.sql(query)
#     metadata = {
#         "path": path,
#         "format": "mysql",
#         "timestamp": dt.datetime.now().timestamp(),
#         "query": query,
#     }
#     return table, metadata



from ..base import BaseDatasetLoader, BaseFileLoader


class CSVFileLoader(BaseFileLoader):
    """CSV file loader.

    This class is responsible for loading CSV files into several dataframe formats,
    duckdb and datafusion.

    Examples:
        ```python
        loader = CSVFileLoader("data.csv")
        df = loader.to_pandas()
    ```
    """

    def model_post_init(self, __context):
        super().model_post_init(__context)
        self.format = "csv"


class CSVDatasetLoader(BaseDatasetLoader):
    """CSV dataset loader.

    This class is responsible for loading CSV files into several dataframe formats,
    duckdb and datafusion.

    Examples:
        ```python
        loader = CSVDatasetLoader("csv_data/")
        df = loader.to_pandas()
        ```
    """

    def model_post_init(self, __context):
        super().model_post_init(__context)
        self.format = "csv"



import importlib.util

from loguru import logger
from typer import Typer

from .pipeline import add as add_pipeline_
from .pipeline import add_job as add_pipeline_job_
from .pipeline import all_pipelines as all_pipelines_
from .pipeline import delete as delete_pipeline_
from .pipeline import get_summary as get_pipeline_summary_
from .pipeline import new as new_pipeline_
from .pipeline import run as run_pipeline_
from .pipeline import run_job as run_pipeline_job_
from .pipeline import save_dag as save_pipeline_dag_
from .pipeline import schedule as schedule_pipeline_
from .pipeline import show_dag as show_pipeline_dag_
from .pipeline import show_summary as show_pipeline_summary_
from .pipeline import start_mqtt_listener as start_mqtt_listener_

if importlib.util.find_spec("apscheduler"):
    from .scheduler import get_schedule_manager
    from .scheduler import start_worker as start_worker_
else:
    get_schedule_manager = None
    start_scheduler_ = None

from . import init as init_

app = Typer()


@app.command()
def run_pipeline(
    pipeline_name: str,
    executor: str = "local",
    base_dir: str = None,
    inputs: str = None,
    final_vars: str = None,
    with_tracker: bool = False,
    reload: bool = False,
    storage_options: str = None,
):
    """
    Run the specified task.
    Args:
        pipeline_name (str): The name of the task.
        executor (str, optional): The executor to use for running the task. Defaults to "local".
        base_dir (str, optional): The base path for the task. Defaults to None.
        inputs (str, optional): The inputs for the task. Defaults to None.
        final_vars (str, optional): The final variables for the task. Defaults to None.
        with_tracker (bool, optional): Whether to use a tracker for the task. Defaults to False.
        reload (bool, optional): Whether to reload the task. Defaults to False.
        storage_options (str, optional): The filesystem storage options for the task. Defaults to None.
    """
    inputs = eval(inputs) if inputs is not None else None
    final_vars = eval(final_vars) if final_vars is not None else None
    with_tracker = with_tracker if with_tracker is not None else None
    storage_options = eval(storage_options) if storage_options is not None else {}

    _ = run_pipeline_(
        name=pipeline_name,
        executor=executor,
        base_dir=base_dir,
        inputs=inputs,
        final_vars=final_vars,
        with_tracker=with_tracker,
        reload=reload,
        storage_options=storage_options,
    )


@app.command()
def run_pipeline_job(
    pipeline_name: str,
    executor: str = "local",
    base_dir: str = None,
    inputs: str = None,
    final_vars: str = None,
    with_tracker: bool = False,
    reload: bool = False,
    storage_options: str = None,
):
    """
    Add a job to run the pipeline with the given parameters to the scheduler.
    Executes the job immediatly.

    Args:
        name (str): The name of the job.
        executor (str, optional): The executor to use for the job. Defaults to None.
        inputs (str, optional): The inputs for the job. Defaults to None.
        final_vars (str, optional): The final variables for the job. Defaults to None.
        with_tracker (bool, optional): Whether to use a tracker for the job. Defaults to None.
        base_dir (str, optional): The base path for the job. Defaults to None.
        reload (bool): Whether to reload the job. Defaults to False.
        storage_options (str, optional): The filesystem storage options for the task. Defaults to None.
    """

    inputs = eval(inputs) if inputs else None
    final_vars = eval(final_vars) if final_vars is not None else None
    with_tracker = with_tracker if with_tracker is not None else None
    storage_options = eval(storage_options) if storage_options is not None else {}

    _ = run_pipeline_job_(
        name=pipeline_name,
        executor=executor,
        base_dir=base_dir,
        inputs=inputs,
        final_vars=final_vars,
        with_tracker=with_tracker,
        reload=reload,
        storage_options=storage_options,
    )


@app.command()
def add_pipeline_job(
    pipeline_name: str,
    executor: str = "local",
    base_dir: str = None,
    inputs: str = None,
    final_vars: str = None,
    with_tracker: bool = False,
    reload: bool = False,
    storage_options: str = None,
):
    """
    Add a job to run the pipeline with the given parameters to the scheduler data store.
    Executes the job immediatly and returns the job id (UUID). The job result will be stored in the data store for the
    given `result_expiration_time` and can be fetched using the job id (UUID).

    Args:
        pipeline_name (str): The name of the job.
        executor (str, optional): The executor to use for the job. Defaults to None.
        inputs (str, optional): The inputs for the job. Defaults to None.
        final_vars (str, optional): The final variables for the job. Defaults to None.
        with_tracker (bool, optional): Whether to use a tracker for the job. Defaults to None.
        base_dir (str, optional): The base path for the job. Defaults to None.
        reload (bool): Whether to reload the job. Defaults to False.
        storage_options (str, optional): The filesystem storage options for the task. Defaults to None.
    """

    inputs = eval(inputs) if inputs else None
    final_vars = eval(final_vars) if final_vars is not None else None
    with_tracker = with_tracker if with_tracker is not None else None
    storage_options = eval(storage_options) if storage_options is not None else {}

    id_ = add_pipeline_job_(
        name=pipeline_name,
        executor=executor,
        base_dir=base_dir,
        inputs=inputs,
        final_vars=final_vars,
        with_tracker=with_tracker,
        reload=reload,
        storage_options=storage_options,
    )
    logger.info(f"Job {id_} added to the scheduler.")


@app.command()
def schedule_pipeline(
    pipeline_name: str,
    executor: str = "local",
    base_dir: str = None,
    type: str = "cron",
    inputs: str = None,
    final_vars: str = None,
    with_tracker: bool = False,
    paused: bool = False,
    coalesce: str = "latest",
    misfire_grace_time: float = None,
    max_jitter: float = None,
    max_running_jobs: int = None,
    conflict_policy: str = "do_nothing",
    crontab: str = None,
    cron_params: str = None,
    interval_params: str = None,
    calendarinterval_params: str = None,
    date_params: str = None,
    storage_options: str = None,
):
    """
    Schedule a job with the given parameters.

    Args:
        pipeline_name (str): The name of the job.
        executor (str, optional): The executor to use for running the job. Defaults to "local".
        base_dir (str, optional): The base path for the job. Defaults to None.
        type (str, optional): The type of the job. Defaults to "cron".
        inputs (str, optional): The inputs for the job. Defaults to None.
        final_vars (str, optional): The final variables for the job. Defaults to None.
        with_tracker (bool, optional): Whether to use a tracker for the job. Defaults to False.
        paused (bool, optional): Whether the job should be initially paused. Defaults to False.
        coalesce (str, optional): The coalesce strategy for the job. Defaults to "latest".
        misfire_grace_time (float, optional): The misfire grace time for the job. Defaults to None.
        max_jitter (float, optional): The maximum jitter for the job. Defaults to None.
        max_running_jobs (int, optional): The maximum number of running jobs. Defaults to None.
        conflict_policy (str, optional): The conflict policy for the job. Defaults to "do_nothing".
        crontab (str, optional): The crontab expression for the job. Defaults to None.
        cron_params (str, optional): Additional parameters for the cron job. Defaults to None.
        interval_params (str, optional): Additional parameters for the interval job. Defaults to None.
        calendarinterval_params (str, optional): Additional parameters for the calendar interval job. Defaults to None.
        date_params (str, optional): Additional parameters for the date job. Defaults to None.
        storage_options (str, optional): The filesystem storage options for the task. Defaults to None.
    """
    if get_schedule_manager is None:
        raise ValueError("APScheduler not installed. Please install it first.")

    inputs = eval(inputs) if inputs else None
    final_vars = eval(final_vars) if final_vars is not None else None
    with_tracker = with_tracker if with_tracker is not None else None
    storage_options = eval(storage_options) if storage_options is not None else {}

    crontab = crontab or None
    cron_params = (
        dict([kw.split("=") for kw in cron_params.split(",")]) if cron_params else {}
    )
    interval_params = (
        dict([kw.split("=") for kw in interval_params.split(",")])
        if interval_params
        else {}
    )

    calendarinterval_params = (
        dict([kw.split("=") for kw in calendarinterval_params.split(",")])
        if calendarinterval_params
        else {}
    )
    date_params = (
        dict([kw.split("=") for kw in date_params.split(",")]) if date_params else {}
    )
    try:
        for key in ["weeks", "days", "hours", "minutes", "seconds"]:
            if key in interval_params:
                interval_params[key] = float(interval_params[key])
            if key in calendarinterval_params:
                calendarinterval_params[key] = float(calendarinterval_params[key])
    except ValueError:
        pass
    kwargs = {
        **cron_params,
        **interval_params,
        **calendarinterval_params,
        **date_params,
    }
    if crontab is not None:
        kwargs["crontab"] = crontab

    id_ = schedule_pipeline_(
        name=pipeline_name,
        executor=executor,
        base_dir=base_dir,
        type=type,
        inputs=inputs,
        final_vars=final_vars,
        with_tracker=with_tracker,
        paused=paused,
        coalesce=coalesce,
        misfire_grace_time=misfire_grace_time,
        max_jitter=max_jitter,
        max_running_jobs=max_running_jobs,
        conflict_policy=conflict_policy,
        storage_options=storage_options,
        **kwargs,
    )
    logger.info(f"Job {id_} scheduled.")


@app.command()
def new_pipeline(
    pipeline_name: str,
    base_dir: str = None,
    overwrite: bool = False,
    storage_options: str = None,
):
    """
    Create a new pipeline with the given parameters.

    Args:
        pipeline_name (str): The name of the pipeline.
        base_dir (str, optional): The base path for the pipeline. Defaults to None.
        overwrite (bool, optional): Whether to overwrite an existing pipeline with the same name. Defaults to False.
        storage_options (str, optional): The filesystem storage options for the task. Defaults to None".
    """

    storage_options = eval(storage_options) if storage_options is not None else {}

    new_pipeline_(
        name=pipeline_name,
        base_dir=base_dir,
        overwrite=overwrite,
        storage_options=storage_options,
    )


@app.command()
def add_pipeline(
    pipeline_name: str,
    base_dir: str = None,
    overwrite: bool = False,
    pipeline_file: str = "",
    pipeline_config: str = "",
    storage_options: str = None,
):
    """
    Create a new pipeline with the given parameters.

    Args:
        pipeline_name (str): The name of the pipeline.
        base_dir (str, optional): The base path for the pipeline. Defaults to None.
        overwrite (bool, optional): Whether to overwrite an existing pipeline with the same name. Defaults to False.
        pipeline_file (str, optional): The path to the pipeline file. Defaults to "".
        pipeline_config (str, optional): The path to the pipeline configuration. Defaults to "".
        storage_options (str, optional): The filesystem storage options for the task. Defaults to None".
    """

    storage_options = eval(storage_options) if storage_options is not None else {}

    add_pipeline_(
        name=pipeline_name,
        base_dir=base_dir,
        overwrite=overwrite,
        pipeline_config=pipeline_config,
        pipeline_file=pipeline_file,
        storage_options=storage_options,
    )


@app.command()
def delete_pipeline(
    pipeline_name: str,
    base_dir: str = None,
    module: bool = False,
    storage_options: str = "{}",
):
    """
    Delete a pipeline.

    Args:
        pipeline_name (str): The name of the pipeline to delete.
        base_dir (str): The base path of the pipeline. Defaults to None.
        module (bool, optional): Whether to delete the pipeline module. Defaults to False.
        storage_options (str, optional): The filesystem storage options for the task. Defaults to None".
    """

    storage_options = eval(storage_options) if storage_options is not None else {}

    delete_pipeline_(
        name=pipeline_name,
        base_dir=base_dir,
        remove_module=module,
        storage_options=storage_options,
    )


@app.command()
def init(
    project_name: str = None,
    base_dir: str = None,
    storage_options: str = None,
):
    """
    Initialize the FlowerPower application.

    Args:
        name (str): The name of the application.
        base_dir (str, optional): The base path of the application. Defaults to "".
        storage_options (str, optional): The filesystem storage options for the task. Defaults to None".
    """
    storage_options = eval(storage_options) if storage_options is not None else {}

    init_(
        name=project_name,
        base_dir=base_dir,
        storage_options=storage_options,
    )


@app.command()
def start_worker(worker_name: str, base_dir: str = None, storage_options: str = "{}"):
    """
    Start a worker.

    Args:
        name (str): The name of the worker.
        base_dir (str, optional): The base path. Defaults to "".
        storage_options (str, optional): The filesystem storage options for the task. Defaults to None".
    """
    storage_options = eval(storage_options) if storage_options is not None else {}

    start_worker_(
        name=worker_name,
        base_dir=base_dir,
        background=False,
        storage_options=storage_options,
    )


@app.command()
def save_pipeline_dag(
    pipeline_name: str,
    base_dir: str = None,
    format: str = "png",
    storage_options: str = "{}",
):
    """
    Show the pipeline.

    Args:
        pipeline_name (str): The name of the pipeline.
        base_dir (str, optional): The base path of the pipeline. Defaults to "".
        storage_options (str, optional): The filesystem storage options for the task. Defaults to None".
    """
    storage_options = eval(storage_options) if storage_options is not None else {}
    save_pipeline_dag_(
        name=pipeline_name,
        base_dir=base_dir,
        format=format,
        storage_options=storage_options,
    )


@app.command()
def show_pipeline_dag(
    pipeline_name: str,
    base_dir: str = None,
    storage_options: str = "{}",
):
    """
    Show the pipeline.

    Args:
        pipeline_name (str): The name of the pipeline.
        base_dir (str, optional): The base path of the pipeline. Defaults to "".
        storage_options (str, optional): The filesystem storage options for the task. Defaults to None".
    """
    storage_options = eval(storage_options) if storage_options is not None else {}
    show_pipeline_dag_(
        name=pipeline_name,
        base_dir=base_dir,
        storage_options=storage_options,
    )


@app.command()
def all_pipelines(base_dir: str = None, storage_options: str = "{}"):
    """
    List all pipelines.

    Args:
        base_dir (str, optional): The base path of the pipelines. Defaults to "".
        storage_options (str, optional): The filesystem storage options for the task. Defaults to None".
    """
    storage_options = eval(storage_options) if storage_options is not None else {}
    all_pipelines_(base_dir=base_dir, storage_options=storage_options)


@app.command()
def get_pipeline_summary(
    pipeline_name: str = None,
    config: bool = True,
    module: bool = True,
    base_dir: str = None,
    storage_options: str = "{}",
):
    """
    Get a summary of the pipeline.

    Args:
        pipeline_name (str, optional): The name of the pipeline.
        show (bool, optional): Whether to show the summary. Defaults to True.
        base_dir (str, optional): The base path of the pipeline. Defaults to "".
        storage_options (str, optional): The filesystem storage options for the task. Defaults to None".
    """
    storage_options = eval(storage_options) if storage_options is not None else {}
    summary = get_pipeline_summary_(
        name=pipeline_name,
        config=config,
        module=module,
        base_dir=base_dir,
        storage_options=storage_options,
    )
    return summary


@app.command()
def show_pipeline_summary(
    pipeline_name: str = None,
    config: bool = True,
    module: bool = True,
    base_dir: str = None,
    storage_options: str = "{}",
):
    """
    Get a summary of the pipeline.

    Args:
        pipeline_name (str, optional): The name of the pipeline.
        show (bool, optional): Whether to show the summary. Defaults to True.
        base_dir (str, optional): The base path of the pipeline. Defaults to "".
        storage_options (str, optional): The filesystem storage options for the task. Defaults to None".
    """
    storage_options = eval(storage_options) if storage_options is not None else {}
    summary = show_pipeline_summary_(
        name=pipeline_name,
        config=config,
        module=module,
        base_dir=base_dir,
        storage_options=storage_options,
    )
    return summary


@app.command()
def start_mqtt_listener(
    pipeline_name: str,
    topic: str = "#",
    host: str = "localhost",
    port: int = 1883,
    user: str = None,
    pw: str = None,
    inputs: str = None,
    final_vars: str = None,
    executor: str = None,
    with_tracker: bool = False,
    with_opentelemetry: bool = False,
    base_dir: str = None,
    storage_options: str = None,
    reload: bool = False,
    result_expiration_time: float = 0.0,
    as_job: bool = False,
    background: bool = False,
):
    """
    Start the MQTT listener.

    Args:
        pipeline_name (str): The name of the pipeline to run.
        topic (str, optional): The MQTT topic to listen to. Defaults to "#".
        host (str, optional): The MQTT host. Defaults to "localhost".
        port (int, optional): The MQTT port. Defaults to 1883.
        user (str, optional): The MQTT username. Defaults to None.
        pw (str, optional): The MQTT password. Defaults to None.
        inputs (str, optional): The inputs for the pipeline. Defaults to None.
        final_vars (list, optional): The final variables for the pipeline. Defaults to None.
        executor (str, optional): The executor to use for the pipeline. Defaults to None.
        with_tracker (bool, optional): Whether to use a tracker for the pipeline. Defaults to False.
        with_opentelemetry (bool, optional): Whether to use OpenTelemetry for the pipeline. Defaults to False.
        base_dir (str, optional): The base path for the pipeline. Defaults to None.
        storage_options (str, optional): The filesystem storage options for the task. Defaults to None.
        reload (bool, optional): Whether to reload the pipeline. Defaults to False.
        result_expiration_time (float, optional): The result expiration time for the pipeline. Defaults to 0.0.
        as_job (bool, optional): Whether to run the pipeline as a job. Defaults to False.
        background (bool, optional): Whether to run the listener in the background. Defaults to e.
    """
    storage_options = eval(storage_options) if storage_options is not None else {}
    start_mqtt_listener_(
        name=pipeline_name,
        topic=topic,
        host=host,
        port=port,
        user=user,
        pw=pw,
        inputs=inputs,
        final_vars=final_vars,
        executor=executor,
        with_tracker=with_tracker,
        with_opentelemetry=with_opentelemetry,
        base_dir=base_dir,
        storage_options=storage_options,
        reload=reload,
        result_expiration_time=result_expiration_time,
        as_job=as_job,
        background=background,
    )


@app.command()
def hamilton_ui(
    port: int = 8241,
    base_dir: str = "~/.hamilton/db",
    no_migration: bool = False,
    no_open: bool = False,
    settings_file: str = "mini",
    config_file: str = None,
):
    """
    Start the Hamilton UI.

    Args:
        port (int, optional): The port to run the UI on. Defaults to 8241.
        base_dir (str, optional): The base path for the UI. Defaults to "~/.hamilton/db".
        no_migration (bool, optional): Whether to run the migration. Defaults to False.
        no_open (bool, optional): Whether to open the UI in the browser. Defaults to False.
        settings_file (str, optional): The settings file to use. Defaults to "mini".
        config_file (str, optional): The config file to use. Defaults to None.
    """
    try:
        from hamilton_ui import commands
    except ImportError:
        logger.error(
            "hamilton[ui] not installed -- you have to install this to run the UI. "
            'Run `pip install "sf-hamilton[ui]"` to install and get started with the UI!'
        )
        raise app.Exit(code=1)

    commands.run(
        port=port,
        base_dir=base_dir,
        no_migration=no_migration,
        no_open=no_open,
        settings_file=settings_file,
        config_file=config_file,
    )


if __name__ == "__main__":
    app()



import uuid

import dill
from sanic import Blueprint
from sanic.exceptions import SanicException
from sanic.response import json, raw

# from sanic_ext import validate


bp = Blueprint("api_flowerpower_scheduler", url_prefix="api/scheduler")


@bp.get("/status")
async def status(request) -> json:
    try:
        status = request.app.ctx.scheduler.state
        return json({"status": status.__str__()})
    except Exception as e:
        raise SanicException(str(e))


@bp.get("/jobs")
async def jobs(request) -> json:
    try:
        jobs = request.app.ctx.scheduler.get_jobs(as_dict=True)
        return json({"jobs": jobs})
    except Exception as e:
        raise SanicException(str(e))


@bp.get("/job-result/<job_id>")
async def job_result(request, job_id: str) -> raw:
    try:
        job_id = uuid.UUID(job_id)
        if job_id not in [job.id for job in request.app.ctx.scheduler.get_jobs()]:
            raise SanicException("Job not found", status_code=404)
        job = request.app.ctx.scheduler.get_job_result(job_id)
        return raw(dill.dumps(job))
    except ValueError:
        raise SanicException("Invalid job ID format", status_code=400)
    except Exception as e:
        raise SanicException(str(e))


@bp.get("/schedules")
async def schedules(request) -> json:
    try:
        schedules = request.app.ctx.scheduler.get_schedules(as_dict=True)
        return json({"schedules": schedules})
    except Exception as e:
        raise SanicException(str(e))


@bp.get("/schedule/<schedule_id>")
async def schedule(request, schedule_id: str) -> json:
    try:
        if schedule_id not in [s.id for s in request.app.ctx.scheduler.get_schedules()]:
            raise SanicException("Schedule not found", status_code=404)
        schedule = request.app.ctx.scheduler.get_schedule(schedule_id, as_dict=True)
        return json(schedule)
    except Exception as e:
        raise SanicException(str(e))


# @bp.post("/schedule")
# async def add_schedule(request) -> json:
#     try:
#         schedule_id = request.app.ctx.scheduler.add_schedule(**body.model_dump())
#         return json({"schedule_id": str(schedule_id)})
#     except Exception as e:
#         raise SanicException(str(e))


# @bp.patch("/schedule/<schedule_id>")
# @openapi.summary("Modify schedule")
# @openapi.description("Modify an existing pipeline schedule")
# @openapi.parameter("schedule_id", str, required=True)
# @openapi.body({"application/json": SchedulerModify}, required=True)
# @openapi.response(200, {"application/json": dict})
# @openapi.response(404, {"application/json": dict})
# async def modify_schedule(request, schedule_id: str) -> json:
#     try:
#         if schedule_id not in [s.id for s in request.app.ctx.scheduler.get_schedules()]:
#             raise SanicException("Schedule not found", status_code=404)

#         body = await deserialize_and_validate(SchedulerModify, body=request.json)
#         request.app.ctx.scheduler.modify_schedule(schedule_id, **body.model_dump())
#         return json({"status": "success"})
#     except Exception as e:
#         raise SanicException(str(e))


@bp.delete("/schedule/<schedule_id>")
async def remove_schedule(request, schedule_id: str) -> json:
    try:
        # body = await deserialize_and_validate(SchedulerDelete, body=request.json)
        if schedule_id not in [s.id for s in request.app.ctx.scheduler.get_schedules()]:
            raise SanicException("Schedule not found", status_code=404)
        request.app.ctx.scheduler.remove_schedule(schedule_id)
        return json({"status": f"deleted schedule with id: {schedule_id}"})
    except Exception as e:
        raise SanicException(str(e))


@bp.delete("/schedules")
async def remove_schedules(request) -> json:
    try:
        request.app.ctx.scheduler.remove_all_schedules()
        return json({"status": "Removed all schedules"})
    except Exception as e:
        raise SanicException(str(e))


@bp.post("/pause-schedule/<schedule_id>")
async def pause_schedule(request, schedule_id: str) -> json:
    try:
        if schedule_id not in [s.id for s in request.app.ctx.scheduler.get_schedules()]:
            raise SanicException("Schedule not found", status_code=404)
        request.app.ctx.scheduler.pause_schedule(schedule_id)
        return json({"status": f"Paused schedule with id: {schedule_id}"})
    except Exception as e:
        raise SanicException(str(e))


# @bp.post("/resume-schedule/<schedule_id>")
# async def resume_schedule(request, schedule_id: str) -> json:
#     try:
#         if schedule_id not in [s.id for s in request.app.ctx.scheduler.get_schedules()]:
#             raise SanicException("Schedule not found", status_code=404)
#         request.app.ctx.scheduler.resume_schedule(schedule_id)
#         return json({"status": "success"})
#     except Exception as e:
#         raise SanicException(str(e))


@bp.get("/tasks")
async def tasks(request) -> json:
    try:
        tasks = request.app.ctx.scheduler.get_tasks(as_dict=True)
        return json({"tasks": tasks})
    except Exception as e:
        raise SanicException(str(e))
